package cz.muni.fi.capella.converter;

import java.io.File;

import org.eclipse.elk.core.ui.views.LayoutPropertySource;
import org.eclipse.elk.core.ui.views.LayoutPropertySourceProvider;
import org.eclipse.elk.core.ui.views.LayoutViewPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.DiagramEditPart;
import org.eclipse.gmf.runtime.diagram.ui.parts.DiagramEditor;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;

/**
 * This class provides necessary constants and utility methods for property import end export.
 * 
 * @author Andrea Jonásová
 * 
 */
public class PropertyUtil {
	
	/**
	 * Prefix of all property files (files that store values for algorithm settings).
	 */
	public static final String PROPERTY_FILE_PREFIX = "alg_"; //$NON-NLS-1$
	
	/**
	 * Sufix (type) of all property files.
	 */
	public static final String PROPERTY_FILE_SUFIX = ".properties"; //$NON-NLS-1$
	
	/**
	 * Path to default property files that are stored in 'preSets' folder in plugin.
	 */
	public static final String PROP_DIRECTORY_PATH = "platform:/plugin/cz.muni.fi.capella.converter.ui/preSets"; //$NON-NLS-1$
	
	/**
	 * Identificator for custom directory setting in preferences. 
	 */
	public static final String CUSTOM_PROPERTY_DIR_PREF_ID = "PATH"; //$NON-NLS-1$
	
	/**
	 * Initial directory in custom directory setting in preferences.
	 */
	public static final String DEFAULT_PROPERTY_DIR = 
			"C:" + File.separator + 
			"Users" + File.separator + 
			"Public" + File.separator + 
			"Documents"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
	
	/**
	 * Gets property store, which is used for exporting and importing values of algorithm settings.
	 * 
	 * @return A property source for layout options.
	 */
	public static LayoutPropertySource getPropertyStore() {
		IWorkbenchPart workbenchPart = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActivePart();

		LayoutViewPart layouViewPart = (LayoutViewPart) workbenchPart;
		LayoutPropertySourceProvider propertySource = layouViewPart.getPropertySourceProvider();
		if (propertySource == null) {
			return null;
		}
		DiagramEditPart editPart = ((DiagramEditor) propertySource.getWorkbenchPart()).getDiagramEditPart();
		LayoutPropertySource store = (LayoutPropertySource) propertySource.getPropertySource(editPart);
		
		return store;
	}
}
