package cz.muni.fi.capella.converter.connector;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.Optional;

import org.eclipse.draw2d.Animation;
import org.eclipse.draw2d.Connection;
import org.eclipse.draw2d.ConnectionLocator;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.Shape;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Insets;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.PointList;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.elk.alg.layered.options.CrossingMinimizationStrategy;
import org.eclipse.elk.alg.layered.options.LayeredOptions;
import org.eclipse.elk.conn.gmf.ApplyLayoutRequest;
import org.eclipse.elk.conn.gmf.IEditPartFilter;
import org.eclipse.elk.core.math.ElkPadding;
import org.eclipse.elk.core.math.KVector;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.options.EdgeLabelPlacement;
import org.eclipse.elk.core.options.NodeLabelPlacement;
import org.eclipse.elk.core.options.PortConstraints;
import org.eclipse.elk.core.options.PortLabelPlacement;
import org.eclipse.elk.core.options.PortSide;
import org.eclipse.elk.core.options.SizeConstraint;
import org.eclipse.elk.core.service.IDiagramLayoutConnector;
import org.eclipse.elk.core.service.LayoutMapping;
import org.eclipse.elk.core.util.ElkUtil;
import org.eclipse.elk.graph.ElkConnectableShape;
import org.eclipse.elk.graph.ElkEdge;
import org.eclipse.elk.graph.ElkEdgeSection;
import org.eclipse.elk.graph.ElkGraphElement;
import org.eclipse.elk.graph.ElkLabel;
import org.eclipse.elk.graph.ElkNode;
import org.eclipse.elk.graph.ElkPort;
import org.eclipse.elk.graph.properties.IProperty;
import org.eclipse.elk.graph.properties.IPropertyHolder;
import org.eclipse.elk.graph.properties.Property;
import org.eclipse.elk.graph.util.ElkGraphUtil;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.transaction.impl.InternalTransactionalEditingDomain;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.RootEditPart;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.commands.CommandStack;
import org.eclipse.gmf.runtime.diagram.ui.editparts.AbstractBorderItemEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.CompartmentEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.ConnectionEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.DiagramEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.IGraphicalEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.LabelEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.ResizableCompartmentEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.ShapeNodeEditPart;
import org.eclipse.gmf.runtime.diagram.ui.editparts.TopGraphicEditPart;
import org.eclipse.gmf.runtime.diagram.ui.figures.BorderedNodeFigure;
import org.eclipse.gmf.runtime.diagram.ui.figures.ResizableCompartmentFigure;
import org.eclipse.gmf.runtime.diagram.ui.parts.DiagramEditor;
import org.eclipse.gmf.runtime.draw2d.ui.figures.WrappingLabel;
import org.eclipse.gmf.runtime.notation.Node;
import org.eclipse.gmf.runtime.notation.View;
import org.eclipse.sirius.diagram.DDiagramElement;
import org.eclipse.sirius.diagram.DDiagramElementContainer;
import org.eclipse.sirius.diagram.DNode;
import org.eclipse.sirius.diagram.DNodeContainer;
import org.eclipse.sirius.diagram.DNodeList;
import org.eclipse.sirius.diagram.DSemanticDiagram;
import org.eclipse.sirius.diagram.LabelPosition;
import org.eclipse.sirius.diagram.NodeStyle;
import org.eclipse.sirius.diagram.business.api.query.DDiagramElementQuery;
import org.eclipse.sirius.diagram.business.api.query.DNodeQuery;
import org.eclipse.sirius.diagram.business.internal.query.DDiagramElementContainerExperimentalQuery;
import org.eclipse.sirius.diagram.business.internal.query.DNodeContainerExperimentalQuery;
import org.eclipse.sirius.diagram.description.style.Side;
import org.eclipse.sirius.diagram.elk.ElkDiagramLayoutConnector;
import org.eclipse.sirius.diagram.ui.business.api.query.EditPartQuery;
import org.eclipse.sirius.diagram.ui.edit.api.part.AbstractDiagramElementContainerEditPart;
import org.eclipse.sirius.diagram.ui.edit.api.part.IAbstractDiagramNodeEditPart;
import org.eclipse.sirius.diagram.ui.edit.api.part.IDiagramElementEditPart;
import org.eclipse.sirius.diagram.ui.internal.edit.parts.AbstractDNodeContainerCompartmentEditPart;
import org.eclipse.sirius.diagram.ui.internal.edit.parts.AbstractDNodeListCompartmentEditPart;
import org.eclipse.sirius.diagram.ui.internal.edit.parts.DNodeListElementEditPart;
import org.eclipse.sirius.diagram.ui.internal.edit.parts.SiriusDescriptionCompartmentEditPart;
import org.eclipse.sirius.diagram.ui.internal.edit.parts.SiriusNoteEditPart;
import org.eclipse.sirius.diagram.ui.internal.edit.parts.SiriusTextEditPart;
import org.eclipse.sirius.diagram.ui.internal.refresh.GMFHelper;
import org.eclipse.sirius.diagram.ui.tools.api.graphical.edit.styles.IBorderItemOffsets;
import org.eclipse.sirius.ext.gmf.runtime.editparts.GraphicalHelper;
import org.eclipse.sirius.ext.gmf.runtime.gef.ui.figures.AlphaDropShadowBorder;
import org.eclipse.sirius.ext.gmf.runtime.gef.ui.figures.SiriusWrapLabel;
import org.eclipse.sirius.viewpoint.LabelAlignment;
import org.eclipse.sirius.viewpoint.LabelStyle;
import org.eclipse.sirius.viewpoint.Style;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.graphics.Font;
import org.eclipse.ui.IWorkbenchPart;
import com.google.common.collect.BiMap;
import com.google.inject.Inject;
import com.google.inject.Singleton;

/**
 * Copied from {@link ElkDiagramLayoutConnector} and modified to fit this solution.
 */
@Singleton
public class CapellaConnector implements IDiagramLayoutConnector {
	
	/** list of connection edit parts that were found in the diagram. */
    public static final IProperty<List<ConnectionEditPart>> CONNECTIONS = 
            new Property<List<ConnectionEditPart>>("gmf.connections");
    
    /** diagram edit part of the currently layouted diagram. */
    public static final IProperty<DiagramEditPart> DIAGRAM_EDIT_PART = new Property<DiagramEditPart>(
            "gmf.diagramEditPart");
    
    /** the command that applies the transferred layout to the diagram. */
    public static final IProperty<Command> LAYOUT_COMMAND = new Property<Command>(
            "gmf.applyLayoutCommand");
    
    /** the command stack that executes the command. */
    public static final IProperty<CommandStack> COMMAND_STACK = new Property<CommandStack>(
            "gmf.applyLayoutCommandStack");
    
    /** the offset to add for all coordinates. */
    public static final IProperty<KVector> COORDINATE_OFFSET = new Property<KVector>(
            "gmf.coordinateOffset");
    
    /**
     * By default, the layout provider will enlarge a container until it is large enough to contain its children. If
     * this option is set, it won't do so (during the application of the layout: {@link #transferLayout(LayoutMapping)}.
     */
    public static final IProperty<Boolean> NODE_SIZE_FIXED_SIZE = new Property<Boolean>("org.eclipse.sirius.diagram.elk.fixedNodeSize", false, null, null);
    
    @Inject
    private IEditPartFilter editPartFilter;
    
    
    /**
     * Calculates the absolute bounds of the given figure.
     * 
     * @param figure a figure
     * @return the absolute bounds
     */
    public static Rectangle getAbsoluteBounds(final IFigure figure) {
        Rectangle bounds = new Rectangle(figure.getBounds()) {
            static final long serialVersionUID = 1;
            @Override
            public void performScale(final double factor) {
                // don't perform any scaling to avoid distortion by the zoom level
            }
        };
        figure.translateToAbsolute(bounds);
        return bounds;
    }
    
    /**
     * If the workbench part is a diagram editor, returns that. Otherwise, returns {@code null}. This is more or less
     * just a fancy cast.
     *
     * @param workbenchPart
     *            the workbench part to check.
     * @return the workbench part as a diagram editor, or {@code null}.
     */
    protected DiagramEditor getDiagramEditor(final IWorkbenchPart workbenchPart) {
        if (workbenchPart instanceof DiagramEditor) {
            return (DiagramEditor) workbenchPart;
        }
        return null;
    }
    
    public static double getShadowBorderSize(final EditPart editPart) {
        double shadowBorderSize = 0;
        if (editPart instanceof AbstractDiagramElementContainerEditPart && ((AbstractDiagramElementContainerEditPart) editPart).isShadowBorderNeeded()) {
            shadowBorderSize = AlphaDropShadowBorder.getDefaultShadowSize();
        }
        return shadowBorderSize;
    }
    
    /**
     * Get the default dimension for a container edit part..
     * 
     * @param editPart
     *            an edit part
     * @return The default dimension for a container edit part
     */
    public static KVector getDefaultDimension(final AbstractDiagramElementContainerEditPart editPart) {
        Dimension defaultDimension = editPart.getDefaultDimension();
        KVector result = new KVector(defaultDimension.preciseWidth(), defaultDimension.preciseHeight());
        // Remove shadow border size for container: Indeed, the edges and border nodes use the "internal figure bounds"
        // (ie without shadow). So for ELK the shadow size must be removed. It will be added before applying the layout
        // in org.eclipse.sirius.diagram.elk.GmfLayoutEditPolicy.addShapeLayout(GmfLayoutCommand, ElkShape,
        // GraphicalEditPart, double).
        double shadowBorderSize = getShadowBorderSize(editPart);
        return result.sub(shadowBorderSize, shadowBorderSize);
    }

	@Override
	public LayoutMapping buildLayoutGraph(IWorkbenchPart workbenchPart, Object diagramPart) {
		// get the diagram editor part
        DiagramEditor diagramEditor = getDiagramEditor(workbenchPart);
        DiagramEditPart diagramEditPart = diagramEditor.getDiagramEditPart();
        
     // choose the layout root edit part
        IGraphicalEditPart layoutRootPart = null;
        List<ShapeNodeEditPart> selectedParts = null;
        if (diagramPart instanceof ShapeNodeEditPart || diagramPart instanceof DiagramEditPart) {
            layoutRootPart = (IGraphicalEditPart) diagramPart;
        } else if (diagramPart instanceof IGraphicalEditPart) {
            EditPart tgEditPart = ((IGraphicalEditPart) diagramPart).getTopGraphicEditPart();
            if (tgEditPart instanceof ShapeNodeEditPart) {
                layoutRootPart = (IGraphicalEditPart) tgEditPart;
            }
        } else if (diagramPart instanceof Collection) {
            Collection<?> selection = (Collection<?>) diagramPart;
            // determine the layout root part from the selection
            for (Object object : selection) {
                if (object instanceof IGraphicalEditPart) {
                    if (layoutRootPart != null) {
                        EditPart parent = commonParent(layoutRootPart, (IGraphicalEditPart) object);
                        if (parent != null && !(parent instanceof RootEditPart)) {
                            layoutRootPart = (IGraphicalEditPart) parent;
                        }
                    } else if (!(object instanceof ConnectionEditPart)) {
                            layoutRootPart = (IGraphicalEditPart) object;
                    }
                }
            }
            // build a list of edit parts that shall be layouted completely
            if (layoutRootPart != null) {
                selectedParts = new ArrayList<ShapeNodeEditPart>(selection.size());
                for (Object object : selection) {
                    if (object instanceof IGraphicalEditPart) {
                        IGraphicalEditPart editPart = (IGraphicalEditPart) object;
                        while (editPart != null && getTopGraphicParentEditPartIfPresent(editPart) != layoutRootPart) {
                            editPart = getTopGraphicParentEditPartIfPresent(editPart);
                        }
                        if (editPart instanceof ShapeNodeEditPart && editPartFilter.filter(editPart) && !selectedParts.contains(editPart)) {
                            if (editPart instanceof SiriusNoteEditPart) {
                                if (new EditPartQuery((SiriusNoteEditPart) editPart).isMovableByAutomaticLayout(Collections.emptyList())) {
                                    selectedParts.add((ShapeNodeEditPart) editPart);
                                }
                            } else {
                                selectedParts.add((ShapeNodeEditPart) editPart);
                            }
                        }
                    }
                }
            }
        }

        if (layoutRootPart == null) {
            layoutRootPart = diagramEditPart;
        }
        
        if (layoutRootPart == null) {
            throw new IllegalArgumentException(
                    "Not supported by this layout connector: Workbench part " + workbenchPart
                            + ", Edit part " + diagramPart);
        }

        // create the mapping
        LayoutMapping mapping = buildLayoutGraph(layoutRootPart, selectedParts, diagramEditPart);

        return mapping;
	}
	
    /**
     * Determine the lowest common parent of the two edit parts.
     * 
     * @param editPart1
     *            the first edit part
     * @param editPart2
     *            the second edit part
     * @return the common parent, or {@code null} if there is none
     */
    protected static EditPart commonParent(final IGraphicalEditPart editPart1, final IGraphicalEditPart editPart2) {
        IGraphicalEditPart ep1 = editPart1;
        IGraphicalEditPart ep2 = editPart2;
        do {
            if (isParent(ep1, ep2)) {
                return ep1;
            }
            if (isParent(ep2, ep1)) {
                return ep2;
            }
            ep1 = getTopGraphicParentEditPartIfPresent(ep1);
            ep2 = getTopGraphicParentEditPartIfPresent(ep2);
        } while (ep1 != null && ep2 != null);
        return null;
    }
    
    /**
     * Return the {@link TopGraphicEditPart} of the parent, if it has one, or the parent itself. This avoids to consider
     * potential intermediate AbstractDNode*CompartmentEditPart as parent.
     * 
     * @return The {@link TopGraphicEditPart} of the parent, if it has one, or the parent itself.
     */
    private static IGraphicalEditPart getTopGraphicParentEditPartIfPresent(IGraphicalEditPart node) {
        EditPart parentEditPart = node.getParent();
        if (parentEditPart instanceof IGraphicalEditPart) {
            IGraphicalEditPart graphicalParentEditPart = (IGraphicalEditPart) parentEditPart;
            TopGraphicEditPart topGraphicEditPart = graphicalParentEditPart.getTopGraphicEditPart();
            return topGraphicEditPart != null ? topGraphicEditPart : graphicalParentEditPart;
        }
        return null;
    }
    
    /**
     * Determine whether the first edit part is a parent of or equals the second one.
     * 
     * @param parent the tentative parent
     * @param child the tentative child
     * @return true if the parent is actually a parent of the child
     */
    protected static boolean isParent(final EditPart parent, final EditPart child) {
        EditPart editPart = child;
        do {
            if (editPart == parent) {
                return true;
            }
            editPart = editPart.getParent();
        } while (editPart != null);
        return false;
    }
    
    /**
     * Creates the actual mapping given an edit part which functions as the root for the layout.
     * 
     * @param layoutRootPart
     *            the layout root edit part
     * @param selection
     *            a selection of contained edit parts to process, or {@code null} if the whole content shall be
     *            processed
     * @param diagramEditPart
     *            the diagram edit part, or {@code null}
     *            
     * @return a layout graph mapping
     */
    protected LayoutMapping buildLayoutGraph(final IGraphicalEditPart layoutRootPart, final List<ShapeNodeEditPart> selection, final DiagramEditPart diagramEditPart) {

        LayoutMapping mapping = new LayoutMapping(null);
        mapping.setProperty(CONNECTIONS, new LinkedList<ConnectionEditPart>());
        mapping.setParentElement(layoutRootPart);

        // find the diagram edit part
        mapping.setProperty(DIAGRAM_EDIT_PART, diagramEditPart);

        ElkNode topNode = ElkGraphUtil.createGraph();
        ElkNode parentNode = topNode;
        // The parentLocation is used when launching an arrange selection of one container contained in another
        // container
        Point parentLocation = new Point(0, 0);

        if (layoutRootPart instanceof ShapeNodeEditPart) {
            // If the root part is a ShapeNodeEditPart and the selection is empty, this implies an arrange selection on
            // only one element (ie one parent). So we want to keep it at a fixed location. For that we use the bounds
            // of the node to determine the size of the graph. We also set its identifier according to its only child.
            // We can not use directly the node as the graph root (in case of it has bordered node) .
            Rectangle childAbsoluteBounds = GraphicalHelper.getAbsoluteBoundsIn100Percent(layoutRootPart, true);
            IGraphicalEditPart parentEditPart = getTopGraphicParentEditPartIfPresent(layoutRootPart);
            if (parentEditPart != null && !(parentEditPart instanceof DiagramEditPart)) {
                // Compute the parent location origin
                parentLocation = GraphicalHelper.getAbsoluteBoundsIn100Percent(parentEditPart).getTopLeft();
            }
            topNode.setLocation(0, 0);
            topNode.setDimensions(childAbsoluteBounds.preciseX() + childAbsoluteBounds.preciseWidth(), childAbsoluteBounds.preciseY() + childAbsoluteBounds.preciseHeight());

            if (selection.isEmpty()) {
                // We add the node to layout to the selection (that is empty)
                selection.add((ShapeNodeEditPart) layoutRootPart);
            } else {
                if (layoutRootPart == diagramEditPart) {
                    String labelText = diagramEditPart.getDiagramView().getName();
                    if (labelText.length() > 0) {
                        ElkLabel label = ElkGraphUtil.createLabel(topNode);
                        label.setText(labelText);
                    }
                } else {
                    topNode.setLocation(childAbsoluteBounds.x, childAbsoluteBounds.y);
                }
                topNode.setDimensions(childAbsoluteBounds.width, childAbsoluteBounds.height);
                mapping.getGraphMap().put(topNode, layoutRootPart);
                // Fix the size of the container. This option is ignored/unknown by ELK, we use it to store this
                // information. It is used later, in transfertLayout, to ignore the layout of this container as we want
                // to keep the same size and location.
                parentNode.setProperty(NODE_SIZE_FIXED_SIZE, true);
            }
        } else {
            // Arrange all the diagram or arrange sub part of the diagram
            // Start with the whole diagram as root for layout. This top node is there to support some meta data.
            Rectangle rootBounds = layoutRootPart.getFigure().getBounds();
            if (layoutRootPart == diagramEditPart) {
                String labelText = diagramEditPart.getDiagramView().getName();
                if (labelText.length() > 0) {
                    ElkLabel label = ElkGraphUtil.createLabel(topNode);
                    label.setText(labelText);
                }
            } else {
                topNode.setLocation(rootBounds.x, rootBounds.y);
            }
            topNode.setDimensions(rootBounds.width, rootBounds.height);
            mapping.getGraphMap().put(topNode, layoutRootPart);
        }
       
        // Set the identifier for the graph according to the diagram name
        if (((View) diagramEditPart.getModel()).getElement() instanceof DSemanticDiagram) {
            topNode.setIdentifier(((DSemanticDiagram) ((View) diagramEditPart.getModel()).getElement()).getName() + "_graph");
        }
        mapping.setLayoutGraph(topNode);

        if (selection != null && !selection.isEmpty()) {
            // layout only the selected elements
            double minx = Integer.MAX_VALUE;
            double miny = Integer.MAX_VALUE;

            for (ShapeNodeEditPart editPart : selection) {
                ElkNode node = createNode(mapping, editPart, parentNode);
                minx = Math.min(minx, node.getX());
                miny = Math.min(miny, node.getY());
                boolean childrenLayouted = buildLayoutGraphRecursively(mapping, node, editPart);
                if (!childrenLayouted && editPart instanceof AbstractDiagramElementContainerEditPart) {
                    node.setProperty(CoreOptions.NODE_SIZE_MINIMUM, getDefaultDimension((AbstractDiagramElementContainerEditPart) editPart));
                }

            }
            if (layoutRootPart instanceof ShapeNodeEditPart) {
                Dimension topLeftInsets = GMFHelper.getContainerTopLeftInsetsAfterLabel((Node) layoutRootPart.getNotationView(), true);
                if (selection.size() == 1 && selection.get(0).equals(layoutRootPart)) {
                    mapping.setProperty(COORDINATE_OFFSET, new KVector(minx - parentLocation.x(), miny - parentLocation.y()));
                } else {
                    // Use the parent node bounds if the arrange selection concerns sub part of a container during
                    // an arrange at opening
                    mapping.setProperty(COORDINATE_OFFSET, new KVector(parentNode.getX() - parentLocation.x() - topLeftInsets.width, parentNode.getY() - parentLocation.y() - topLeftInsets.height));
                }
            } else {
            	mapping.setProperty(COORDINATE_OFFSET, new KVector(minx - parentLocation.x(), miny - parentLocation.y()));
            }
        } else {
            // traverse all children of the layout root part
            buildLayoutGraphRecursively(mapping, topNode, layoutRootPart);
        }

        // transform all connections in the selected area
        processConnections(mapping);

        return mapping;
    }
    
    /**
     * Transfer all layout data from the last created KGraph instance to the original diagram. The diagram is not
     * modified yet, but all required preparations are performed. This is separated from
     * {@link #applyLayout(LayoutMapping)} to allow better code modularization.
     * 
     * @param mapping
     *            a layout mapping that was created by this layout connector
     */
    public void transferLayout(final LayoutMapping mapping) {
        // create a new request to change the layout
        ApplyLayoutRequest applyLayoutRequest = new ApplyLayoutRequest();
        for (Entry<ElkGraphElement, Object> entry : mapping.getGraphMap().entrySet()) {
            if (!(entry.getValue() instanceof DiagramEditPart)) {
                ElkGraphElement graphElement = entry.getKey();
                IGraphicalEditPart part = (IGraphicalEditPart) entry.getValue();
                if (!(part instanceof AbstractDNodeListCompartmentEditPart || part instanceof AbstractDNodeContainerCompartmentEditPart)) {
                    // We ignore compartment edit part that are created into ELK side just to have good layout results
                    if (!graphElement.getProperty(NODE_SIZE_FIXED_SIZE)) {
                        // We do not modify layout of node/graph if the option "Fixed Graph Size" has been set before.
                        applyLayoutRequest.addElement(graphElement, part);
                    }
                }
            }
        }

        ElkNode layoutGraph = mapping.getLayoutGraph();
        applyLayoutRequest.setUpperBound(layoutGraph.getWidth(), layoutGraph.getHeight());

        // correct the layout by adding the offset determined from the selection
        KVector offset = mapping.getProperty(COORDINATE_OFFSET);
        if (offset != null) {
            addOffset(mapping.getLayoutGraph(), offset);
        }

        // check the validity of the editing domain to catch cases where it is
        // disposed
        DiagramEditPart diagramEditPart = mapping.getProperty(DIAGRAM_EDIT_PART);
        if (((InternalTransactionalEditingDomain) diagramEditPart.getEditingDomain()).getChangeRecorder() != null) {
            // retrieve a command for the request; the command is created by
            // GmfLayoutEditPolicy
            Command applyLayoutCommand = diagramEditPart.getCommand(applyLayoutRequest);
            mapping.setProperty(LAYOUT_COMMAND, applyLayoutCommand);
        }
    }
    
    /**
     * Add the given offset to all direct children of the given graph.
     * 
     * @param parentNode
     *            the parent node
     * @param offset
     *            the offset to add
     */
    protected static void addOffset(final ElkNode parentNode, final KVector offset) {
        // correct the offset with the minimal computed coordinates
        double minx = Integer.MAX_VALUE;
        double miny = Integer.MAX_VALUE;
        for (ElkNode child : parentNode.getChildren()) {
            minx = Math.min(minx, child.getX());
            miny = Math.min(miny, child.getY());
        }

        // add the corrected offset
        offset.add(-minx, -miny);
        ElkUtil.translate(parentNode, offset.x, offset.y);
    }
    
    /**
     * Recursively builds a layout graph by analyzing the children of the given edit part.
     * 
     * @param mapping
     *            the layout mapping
     * @param parentLayoutNode
     *            the {@link ElkNode} corresponding to the parent edit part of the current edit part
     * @param currentEditPart
     *            the currently analyzed edit part
     * @return false if no child has been layouted, true otherwise
     */
    protected boolean buildLayoutGraphRecursively(final LayoutMapping mapping, final ElkNode parentLayoutNode, final IGraphicalEditPart currentEditPart) {
        boolean childrenLayouted = false;
        // iterate through the children of the element
        double maxChildShadowBorderSize = -1;
        for (Object obj : currentEditPart.getChildren()) {

            // check visibility of the child
            if (obj instanceof IGraphicalEditPart) {
                IFigure figure = ((IGraphicalEditPart) obj).getFigure();
                if (!figure.isVisible()) {
                    continue;
                }
            }

            // process a port (border item)
            if (obj instanceof AbstractBorderItemEditPart) {
                AbstractBorderItemEditPart borderItem = (AbstractBorderItemEditPart) obj;
                if (editPartFilter.filter(borderItem)) {
                    createPort(mapping, borderItem, currentEditPart, parentLayoutNode);
                }

                // process a compartment, which may contain other elements
            } else if (obj instanceof ResizableCompartmentEditPart && ((CompartmentEditPart) obj).getChildren().size() > 0) {

                CompartmentEditPart compartment = (CompartmentEditPart) obj;
                if (editPartFilter.filter(compartment)) {
                    boolean compExp = true;
                    IFigure compartmentFigure = compartment.getFigure();
                    if (compartmentFigure instanceof ResizableCompartmentFigure) {
                        ResizableCompartmentFigure resizCompFigure = (ResizableCompartmentFigure) compartmentFigure;
                        // check whether the compartment is collapsed
                        compExp = resizCompFigure.isExpanded();
                    }

                    if (compExp) {
                        ElkNode intermediateNode = parentLayoutNode;
                        if (currentEditPart instanceof IDiagramElementEditPart) {
                            IDiagramElementEditPart ideep = (IDiagramElementEditPart) currentEditPart;
                            DDiagramElement dde = ideep.resolveDiagramElement();
                            if (dde instanceof DNodeList || (dde instanceof DNodeContainer && (new DNodeContainerExperimentalQuery((DNodeContainer) dde)).isHorizontaltackContainer()
                                    || new DNodeContainerExperimentalQuery((DNodeContainer) dde).isVerticalStackContainer())) {
                                childrenLayouted = true;
                                // Create a node representing the compartment
                                intermediateNode = createNode(mapping, compartment, parentLayoutNode);
                                // Add some additional layout option to its container to have "fit" layout result
                                Dimension topLeftInsets = GMFHelper.getContainerTopLeftInsetsAfterLabel((Node) compartment.getNotationView(), true);
                                Dimension borderSize = GMFHelper.getBorderSize((DDiagramElementContainer) dde);
                                int separatorLineHeight = 1;
                                ElkPadding padding;
                                if (dde instanceof DNodeList) {
                                    padding = new ElkPadding(topLeftInsets.preciseHeight() + separatorLineHeight, topLeftInsets.preciseWidth(), borderSize.height(), topLeftInsets.preciseWidth());
                                } else {
                                    // For container with VStack, the label of region occupied all the width so we have
                                    // not to use insets
                                    padding = new ElkPadding(topLeftInsets.preciseHeight(), borderSize.preciseWidth(), borderSize.preciseHeight(), borderSize.preciseWidth());
                                }
                                parentLayoutNode.setProperty(CoreOptions.PADDING, padding);
                                // no space around regions
                                parentLayoutNode.setProperty(CoreOptions.SPACING_NODE_NODE, 0d);
                                // no space around labels
                                parentLayoutNode.setProperty(CoreOptions.NODE_LABELS_PADDING, new ElkPadding());
                                // Strategy set to INTERACTIVE to keep order of children
                                parentLayoutNode.setProperty(LayeredOptions.CROSSING_MINIMIZATION_STRATEGY, CrossingMinimizationStrategy.INTERACTIVE);

                                if (dde instanceof DNodeContainer) {
                                    intermediateNode.setProperty(CoreOptions.PADDING, new ElkPadding(0, 0, 0, 0));
                                    intermediateNode.setProperty(CoreOptions.SPACING_NODE_NODE, 0d);
                                    intermediateNode.setProperty(LayeredOptions.CROSSING_MINIMIZATION_STRATEGY, CrossingMinimizationStrategy.INTERACTIVE);
                                }
                                intermediateNode.setProperty(CoreOptions.NODE_LABELS_PADDING, new ElkPadding());
                            } else if (dde instanceof DNodeContainer) {
                                DNodeContainerExperimentalQuery query = new DNodeContainerExperimentalQuery((DNodeContainer) dde);
                                if (query.isHorizontaltackContainer()) {
                                } else if (query.isVerticalStackContainer()) {
                                    // Add some additional layout option to its container to have "fit" layout result
                                    Dimension topLeftInsets = GMFHelper.getContainerTopLeftInsets((Node) compartment.getNotationView(), true);
                                    ElkPadding padding = new ElkPadding(topLeftInsets.preciseHeight(), topLeftInsets.preciseWidth(), topLeftInsets.preciseHeight(), topLeftInsets.preciseWidth());
                                    parentLayoutNode.setProperty(CoreOptions.PADDING, padding);
                                    // no space around regions
                                    parentLayoutNode.setProperty(CoreOptions.SPACING_NODE_NODE, 0d);
                                    // Strategy set to INTERACTIVE to keep order of children
                                    parentLayoutNode.setProperty(LayeredOptions.CROSSING_MINIMIZATION_STRATEGY, CrossingMinimizationStrategy.INTERACTIVE);
                                }
                            }
                        }
                        childrenLayouted = buildLayoutGraphRecursively(mapping, intermediateNode, compartment) || childrenLayouted;
                    }
                }

                // process a node, which may be a parent of ports, compartments,
                // or other nodes
            } else if (obj instanceof ShapeNodeEditPart) {
                ShapeNodeEditPart childNodeEditPart = (ShapeNodeEditPart) obj;
                if (editPartFilter.filter(childNodeEditPart)) {
                    childrenLayouted = true;
                    ElkNode node = createNode(mapping, childNodeEditPart, parentLayoutNode);
                    maxChildShadowBorderSize = Math.max(maxChildShadowBorderSize, getShadowBorderSize(childNodeEditPart));
                    // process the child as new current edit part
                    boolean currentChildrenLayouted = buildLayoutGraphRecursively(mapping, node, childNodeEditPart);
                    if (!currentChildrenLayouted && childNodeEditPart instanceof AbstractDiagramElementContainerEditPart) {
                        Dimension defaultDimension = ((AbstractDiagramElementContainerEditPart) childNodeEditPart).getDefaultDimension();
                        node.setProperty(CoreOptions.NODE_SIZE_MINIMUM, new KVector(defaultDimension.width(), defaultDimension.height()));
                    }
                }

                // process a label of the current node
            } else if (obj instanceof IGraphicalEditPart) {
                ElkLabel newNodeLabel = createNodeLabel(mapping, (IGraphicalEditPart) obj, currentEditPart, parentLayoutNode);
                if (newNodeLabel != null) {
                    parentLayoutNode.getLabels().add(newNodeLabel);
                }
            }
            if (maxChildShadowBorderSize >= 0 && currentEditPart.getNotationView() instanceof Node && !(currentEditPart instanceof ResizableCompartmentEditPart)) {
                // maxChildShadowBorderSize == 0 : There is at least one child, so we set insets of this container
                // maxChildShadowBorderSize > 0 : There is at least one child with a shadow border, we add this
                // border
                // size to the insets to avoid potential scrollbar appearance during the layout application
                // (org.eclipse.sirius.diagram.elk.GmfLayoutEditPolicy.addShapeLayout(GmfLayoutCommand, ElkShape,
                // GraphicalEditPart, double)).
                Dimension topLeftInsets = GMFHelper.getContainerTopLeftInsets((Node) currentEditPart.getNotationView(), true);
                // We directly reuse the left inset to sets the bottom and right insets.
                ElkPadding ei = new ElkPadding(topLeftInsets.preciseHeight(), topLeftInsets.preciseWidth() + maxChildShadowBorderSize, topLeftInsets.preciseWidth() + maxChildShadowBorderSize,
                        topLeftInsets.preciseWidth());
                parentLayoutNode.setProperty(CoreOptions.PADDING, ei);
            }

        }
        return childrenLayouted;
    }
    
    /**
     * Create a node while building the layout graph.
     * 
     * @param mapping
     *            the layout mapping
     * @param nodeEditPart
     *            the node edit part
     * @param parentElkNode
     *            the corresponding parent layout node
     * @return the created node
     */
    protected ElkNode createNode(final LayoutMapping mapping, final IGraphicalEditPart nodeEditPart, final ElkNode parentElkNode) {

        IFigure nodeFigure = nodeEditPart.getFigure();
        ElkNode newNode = ElkGraphUtil.createNode(parentElkNode);

        // set location and size
        Rectangle childAbsoluteBounds = GraphicalHelper.getAbsoluteBoundsIn100Percent(nodeEditPart, true);
        KVector containerAbsoluteLocation = new KVector();
        ElkUtil.toAbsolute(containerAbsoluteLocation, parentElkNode);
        newNode.setX(childAbsoluteBounds.x - containerAbsoluteLocation.x);
        newNode.setY(childAbsoluteBounds.y - containerAbsoluteLocation.y);
        // Remove shadow border size for container: Indeed, the edges and border nodes use the "internal figure bounds"
        // (ie without shadow). So for ELK the shadow size must be removed. It will be added before applying the layout
        // in org.eclipse.sirius.diagram.elk.GmfLayoutEditPolicy.addShapeLayout(GmfLayoutCommand, ElkShape,
        // GraphicalEditPart, double).
        double shadowBorderSize = getShadowBorderSize(nodeEditPart);
        newNode.setDimensions(childAbsoluteBounds.width - shadowBorderSize, childAbsoluteBounds.height - shadowBorderSize);

        // useful to debug.
        if (nodeEditPart instanceof AbstractDNodeListCompartmentEditPart || nodeEditPart instanceof AbstractDNodeContainerCompartmentEditPart) {
            newNode.setIdentifier("Compartment");
        } else if (((View) nodeEditPart.getModel()).getElement() instanceof DDiagramElement) {
            newNode.setIdentifier(((DDiagramElement) ((View) nodeEditPart.getModel()).getElement()).getName());
        }

        // determine minimal size of the node
        try {
            Dimension minSize = nodeFigure.getMinimumSize();
            newNode.setProperty(CoreOptions.NODE_SIZE_MINIMUM, new KVector(minSize.width, minSize.height));
        } catch (SWTException exception) {
            // getMinimumSize() can cause this exception when fonts are disposed
            // for some reason;
            // ignore exception and leave the default minimal size
        }
        if (nodeEditPart instanceof SiriusNoteEditPart || nodeEditPart instanceof SiriusTextEditPart) {
            // For the Note and Text, we want to have a fixed size (defined by the user)
            newNode.setProperty(CoreOptions.NODE_SIZE_CONSTRAINTS, SizeConstraint.fixed());
        }

        if (parentElkNode != null) {
            parentElkNode.getChildren().add(newNode);
        }
        mapping.getGraphMap().put(newNode, nodeEditPart);

        // Create label for Node, not Container, with name inside the node, not on border
        final EObject eObj = nodeEditPart.resolveSemanticElement();
        if (eObj instanceof DNode && ((NodeStyle) ((DNode) eObj).getStyle()).getLabelPosition() == LabelPosition.NODE_LITERAL) {
            ElkLabel newNodeLabel = createNodeLabel(mapping, nodeEditPart, (IGraphicalEditPart) nodeEditPart.getParent(), parentElkNode);
            if (newNodeLabel != null) {
                newNode.getLabels().add(newNodeLabel);
            }
        }

        // store all the connections to process them later
        addConnections(mapping, nodeEditPart);
        return newNode;
    }
    
    /**
     * Determines the insets for a parent figure, relative to the given child. Subclasses may override this if the
     * generic insets calculation does not work.
     * 
     * @param parent
     *            the figure of a parent edit part
     * @param child
     *            the figure of a child edit part
     * @return the insets to add to the relative coordinates of the child
     */
    protected Insets calcSpecificInsets(final IFigure parent, final IFigure child) {
        Insets result = new Insets(0);
        IFigure currentChild = child;
        IFigure currentParent = child.getParent();
        Point coordsToAdd = null;
        boolean isRelative = false;
        // follow the chain of parents in the figure hierarchy up to the given
        // parent figure
        while (currentChild != parent && currentParent != null) {
            if (currentParent.isCoordinateSystem()) {
                // the content of the current parent is relative to that
                // figure's position
                isRelative = true;
                result.add(currentParent.getInsets());
                if (coordsToAdd != null) {
                    // add the position of the previous parent with local
                    // coordinate system
                    result.left += coordsToAdd.x;
                    result.top += coordsToAdd.y;
                }
                coordsToAdd = currentParent.getBounds().getLocation();
            } else if (currentParent == parent && coordsToAdd != null) {
                // we found the top parent, and it does not have local
                // coordinate system,
                // so subtract the parent's coordinates from the previous
                // parent's position
                Point parentCoords = parent.getBounds().getLocation();
                result.left += coordsToAdd.x - parentCoords.x;
                result.top += coordsToAdd.y - parentCoords.y;
            }
            currentChild = currentParent;
            currentParent = currentChild.getParent();
        }
        if (!isRelative) {
            // there is no local coordinate system, so just subtract the
            // coordinates
            Rectangle parentBounds = parent.getBounds();
            currentParent = child.getParent();
            Rectangle containerBounds = currentParent.getBounds();
            result.left = containerBounds.x - parentBounds.x;
            result.top = containerBounds.y - parentBounds.y;
        }
        // In theory it would be better to get the bottom and right insets from
        // the size.
        // However, due to the unpredictability of Draw2D layout managers, this
        // leads to
        // bad results in many cases, so a fixed insets value is more stable.
        result.right = result.left;
        result.bottom = result.left;
        return result;
    }
    
    /**
     * Create a port while building the layout graph.
     * 
     * @param mapping
     *            the layout mapping
     * @param portEditPart
     *            the port edit part
     * @param nodeEditPart
     *            the parent node edit part
     * @param elknode
     *            the corresponding layout node
     * @return the created port
     */
    protected ElkPort createPort(final LayoutMapping mapping, final AbstractBorderItemEditPart portEditPart, final IGraphicalEditPart nodeEditPart, final ElkNode elknode) {

        ElkPort port = ElkGraphUtil.createPort(elknode);

        // Set the location of port label fixed (indeed the location returned by ELK is currently not compatible with
        // Sirius constraint). The label is, by default, located centered below the port.
        elknode.setProperty(CoreOptions.PORT_LABELS_PLACEMENT, PortLabelPlacement.fixed());

        // set the port's layout, relative to the node position
        Rectangle portBounds = getAbsoluteBounds(portEditPart.getFigure());
        Rectangle nodeBounds = getAbsoluteBounds(nodeEditPart.getFigure());
        double xpos = portBounds.x - nodeBounds.x;
        double ypos = portBounds.y - nodeBounds.y;
        port.setLocation(xpos, ypos);
        port.setDimensions(portBounds.width, portBounds.height);

        // Compute the border node offset from Sirius
        double borderNodeOffset = -IBorderItemOffsets.DEFAULT_OFFSET.preciseWidth();
        final EObject eObj = portEditPart.resolveSemanticElement();
        if (eObj instanceof DDiagramElement) {
            if (new DDiagramElementQuery((DDiagramElement) eObj).isIndirectlyCollapsed()) {
                borderNodeOffset = -IBorderItemOffsets.COLLAPSE_FILTER_OFFSET.preciseWidth();
            }
        }
        port.setProperty(CoreOptions.PORT_BORDER_OFFSET, borderNodeOffset);

        // If the port has location constraints, we add it
        if (eObj instanceof DNode) {
            PortSide authorizedSide = PortSide.UNDEFINED;
            DNodeQuery query = new DNodeQuery((DNode) eObj);
            List<Side> forbiddenSides = query.getForbiddenSide();
            Set<PortSide> authorizedSides = new HashSet<PortSide>(PortSide.SIDES_NORTH_EAST_SOUTH_WEST);
            for (Side side : forbiddenSides) {
                authorizedSides.remove(convertSideToPortSide(side));
            }
            if (authorizedSides.size() != 4) {
                // If the 4 sides are authorizes there is no constraint.
                if (authorizedSides.size() > 1) {
                    // Only consider the first in UI order of Sirius, ie WEST, SOUTH, EAST, NORTH (as ELK constraint
                    // only handles one side)
                    if (authorizedSides.contains(PortSide.WEST)) {
                        authorizedSide = PortSide.WEST;
                    } else if (authorizedSides.contains(PortSide.SOUTH)) {
                        authorizedSide = PortSide.SOUTH;
                    } else if (authorizedSides.contains(PortSide.EAST)) {
                        authorizedSide = PortSide.EAST;
                    } else if (authorizedSides.contains(PortSide.NORTH)) {
                        authorizedSide = PortSide.NORTH;
                    }
                } else {
                    authorizedSide = authorizedSides.iterator().next();
                }
                elknode.setProperty(CoreOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_SIDE);
                port.setProperty(CoreOptions.PORT_SIDE, authorizedSide);
            }
        }

        mapping.getGraphMap().put(port, portEditPart);

        // store all the connections to process them later
        addConnections(mapping, portEditPart);

        // set the port label
        for (Object portChildObj : portEditPart.getChildren()) {
            if (portChildObj instanceof IGraphicalEditPart) {
                IFigure labelFigure = ((IGraphicalEditPart) portChildObj).getFigure();
                String text = null;
                if (labelFigure instanceof WrappingLabel) {
                    text = ((WrappingLabel) labelFigure).getText();
                } else if (labelFigure instanceof Label) {
                    text = ((Label) labelFigure).getText();
                } else if (labelFigure instanceof SiriusWrapLabel) {
                    SiriusWrapLabel label = (SiriusWrapLabel) labelFigure;
                    text = label.getText();
                }

                if (text != null) {
                    ElkLabel portLabel = ElkGraphUtil.createLabel(port);
                    portLabel.setText(text);
                    mapping.getGraphMap().put(portLabel, portChildObj);

                    // Set the port label's layout (centered below the port)
                    Rectangle labelBounds = getAbsoluteBounds(labelFigure);
                    portLabel.setLocation((portBounds.preciseWidth() / 2) - (labelBounds.preciseWidth() / 2), portBounds.height + 1);
                    try {
                        Dimension size = labelFigure.getPreferredSize();
                        portLabel.setDimensions(size.width, size.height);
                    } catch (SWTException exception) {
                        // ignore exception and leave the label size to (0, 0)
                    }

                    // We would set the modified flag to false here, but that
                    // doesn't exist anymore
                }
            }
        }

        return port;
    }
    
    private PortSide convertSideToPortSide(Side side) {
        PortSide result = PortSide.UNDEFINED;
        if (Side.WEST.equals(side)) {
            result = PortSide.WEST;
        } else if (Side.EAST.equals(side)) {
            result = PortSide.EAST;
        } else if (Side.NORTH.equals(side)) {
            result = PortSide.NORTH;
        } else if (Side.SOUTH.equals(side)) {
            result = PortSide.SOUTH;
        }
        return result;
    }
    
    /**
     * Create a node label while building the layout graph.
     * 
     * @param mapping
     *            the layout mapping
     * @param labelEditPart
     *            the label edit part
     * @param nodeEditPart
     *            the parent node edit part
     * @param elknode
     *            the layout node for which the label is set
     * @return the created label
     */
    protected ElkLabel createNodeLabel(final LayoutMapping mapping, final IGraphicalEditPart labelEditPart, final IGraphicalEditPart nodeEditPart, final ElkNode elknode) {

        IFigure labelFigure;
        // We do not handle label of Note or Text (we want a fixed size for them).
        if (!(labelEditPart instanceof SiriusDescriptionCompartmentEditPart)) {
            if (labelEditPart instanceof IAbstractDiagramNodeEditPart) {
                labelFigure = ((IAbstractDiagramNodeEditPart) labelEditPart).getNodeLabel();
            } else {
                labelFigure = labelEditPart.getFigure();
            }
            String text = null;
            Font font = null;

            if (labelFigure instanceof WrappingLabel) {
                WrappingLabel wrappingLabel = (WrappingLabel) labelFigure;
                text = wrappingLabel.getText();
                font = wrappingLabel.getFont();
            } else if (labelFigure instanceof Label) {
                Label label = (Label) labelFigure;
                text = label.getText();
                font = label.getFont();
            } else if (labelFigure instanceof SiriusWrapLabel) {
                SiriusWrapLabel label = (SiriusWrapLabel) labelFigure;
                text = label.getText();
                font = label.getFont();
            }

            if (text != null) {
                ElkLabel label = ElkGraphUtil.createLabel(elknode);
                if (text.isEmpty()) {
                    // Set at least a white space character (otherwise, ELK "ignores" this label and the layout of its
                    // container is not OK (wrong width/height)
                    label.setText(" ");
                } else {
                    label.setText(text);
                }
                Rectangle labelBounds = getAbsoluteBounds(labelFigure);
                Rectangle nodeBounds;
                if (!(labelEditPart instanceof IAbstractDiagramNodeEditPart)) {
                    mapping.getGraphMap().put(label, labelEditPart);
                    nodeBounds = getAbsoluteBounds(nodeEditPart.getFigure());
                } else {
                    nodeBounds = getAbsoluteBounds(labelEditPart.getFigure());
                }
                label.setLocation(labelBounds.x - nodeBounds.x, labelBounds.y - nodeBounds.y);

                try {
                    // We use the preferred size and not the labelBounds to "reset" previous manual wrapping size
                    Dimension size = labelFigure.getPreferredSize();
                    label.setDimensions(size.width, size.height);
                    if (font != null && !font.isDisposed()) {
                        label.setProperty(CoreOptions.FONT_NAME, font.getFontData()[0].getName());
                        label.setProperty(CoreOptions.FONT_SIZE, font.getFontData()[0].getHeight());
                    }
                } catch (SWTException exception) {
                    // ignore exception and leave the label size to (0, 0)
                }

                // Set globally the location of the label according to the style
                NodeLabelPlacement insideLabelPlacement = NodeLabelPlacement.INSIDE;
                NodeLabelPlacement verticalNodeLabelPlacement = NodeLabelPlacement.V_TOP;
                NodeLabelPlacement horizontalLabelPlacement = NodeLabelPlacement.H_CENTER;

                // For ListElement, we force H_CENTER and V_CENTER to have good layout result (fit).
                if (labelEditPart instanceof DNodeListElementEditPart) {
                    verticalNodeLabelPlacement = NodeLabelPlacement.V_CENTER;
                } else {
                    EObject siriusObject;
                    if (labelEditPart instanceof IAbstractDiagramNodeEditPart) {
                        siriusObject = labelEditPart.resolveSemanticElement();
                    } else {
                        siriusObject = nodeEditPart.resolveSemanticElement();
                    }
                    boolean forcedValue = false;
                    if (siriusObject instanceof DDiagramElement) {
                        if (siriusObject instanceof DNodeContainer) {
                            if (new DDiagramElementContainerExperimentalQuery((DNodeContainer) siriusObject).isRegionInVerticalStack()) {
                                // Keep default value (OK for this kind of container)
                                forcedValue = true;
                                verticalNodeLabelPlacement = NodeLabelPlacement.V_CENTER;
                            } else if (new DDiagramElementContainerExperimentalQuery((DNodeContainer) siriusObject).isRegionInHorizontalStack()) {
                                forcedValue = true;
                                // verticalNodeLabelPlacement = NodeLabelPlacement.V_TOP;
                                // horizontalLabelPlacement = NodeLabelPlacement.H_CENTER;
                            }
                        }
                        if (!forcedValue) {
                            DDiagramElement dde = (DDiagramElement) siriusObject;
                            Style style = dde.getStyle();
                            if (style instanceof LabelStyle) {
                                LabelAlignment labelAlignment = ((LabelStyle) style).getLabelAlignment();
                                if (labelAlignment.equals(LabelAlignment.LEFT)) {
                                    horizontalLabelPlacement = NodeLabelPlacement.H_LEFT;
                                } else if (labelAlignment.equals(LabelAlignment.RIGHT)) {
                                    horizontalLabelPlacement = NodeLabelPlacement.H_RIGHT;
                                }
                            }
                            if (style instanceof NodeStyle) {
                                if (((NodeStyle) style).getLabelPosition().equals(LabelPosition.BORDER_LITERAL)) {
                                    insideLabelPlacement = NodeLabelPlacement.OUTSIDE;
                                }
                                verticalNodeLabelPlacement = NodeLabelPlacement.V_CENTER;
                            }
                        }
                    }
                }
                EnumSet<NodeLabelPlacement> enumSet = EnumSet.of(insideLabelPlacement, horizontalLabelPlacement, verticalNodeLabelPlacement);
                label.setProperty(CoreOptions.NODE_LABELS_PLACEMENT, enumSet);

                return label;
            }
        }
        return null;
    }
    
    /**
     * Adds all target connections and connected connections to the list of connections that must be processed later.
     * 
     * @param mapping
     *            the layout mapping
     * @param editPart
     *            an edit part
     */
    protected void addConnections(final LayoutMapping mapping, final IGraphicalEditPart editPart) {
        for (Object targetConn : editPart.getTargetConnections()) {
            if (targetConn instanceof ConnectionEditPart) {
                ConnectionEditPart connectionEditPart = (ConnectionEditPart) targetConn;
                if (editPartFilter.filter(connectionEditPart)) {
                    mapping.getProperty(CONNECTIONS).add(connectionEditPart);
                    addConnections(mapping, connectionEditPart);
                }
            }
        }
    }
    
    /**
     * Creates new edges and takes care of the labels for each connection identified in the
     * {@code buildLayoutGraphRecursively} method.
     * 
     * @param mapping
     *            the layout mapping
     */
    protected void processConnections(final LayoutMapping mapping) {
        Map<EReference, ElkEdge> reference2EdgeMap = new HashMap<>();
        
        for (ConnectionEditPart connection : mapping.getProperty(CONNECTIONS)) {
            boolean isOppositeEdge = false;
            Optional<EdgeLabelPlacement> edgeLabelPlacement = Optional.empty();
            ElkEdge edge;

            // Check whether the edge belongs to an Ecore reference, which may have opposites.
            // This is required for the layout of Ecore diagrams, since the bend points of
            // opposite references are kept synchronized by the editor.
            EObject modelObject = connection.getNotationView().getElement();
            if (modelObject instanceof EReference) {
                EReference reference = (EReference) modelObject;
                edge = reference2EdgeMap.get(reference.getEOpposite());
                if (edge != null) {
                    edgeLabelPlacement = Optional.of(EdgeLabelPlacement.TAIL);
                    isOppositeEdge = true;
                } else {
                    edge = ElkGraphUtil.createEdge(null);
                    reference2EdgeMap.put(reference, edge);
                }
            } else {
                edge = ElkGraphUtil.createEdge(null);
            }

            BiMap<Object, ElkGraphElement> inverseGraphMap = mapping.getGraphMap().inverse();

            // find a proper source node and source port
            ElkGraphElement sourceElem;
            EditPart sourceObj = connection.getSource();
            if (sourceObj instanceof ConnectionEditPart) {
                sourceElem = inverseGraphMap.get(((ConnectionEditPart) sourceObj).getSource());
                if (sourceElem == null) {
                    sourceElem = inverseGraphMap.get(((ConnectionEditPart) sourceObj).getTarget());
                }
            } else {
                sourceElem = inverseGraphMap.get(sourceObj);
            }
            
            ElkConnectableShape sourceShape = null;
            ElkPort sourcePort = null;
            ElkNode sourceNode = null;
            if (sourceElem instanceof ElkNode) {
                sourceNode = (ElkNode) sourceElem;
                sourceShape = sourceNode;
            } else if (sourceElem instanceof ElkPort) {
                sourcePort = (ElkPort) sourceElem;
                sourceNode = sourcePort.getParent();
                sourceShape = sourcePort;
            } else {
                continue;
            }

            // find a proper target node and target port
            ElkGraphElement targetElem;
            EditPart targetObj = connection.getTarget();
            if (targetObj instanceof ConnectionEditPart) {
                targetElem = inverseGraphMap.get(((ConnectionEditPart) targetObj).getTarget());
                if (targetElem == null) {
                    targetElem = inverseGraphMap.get(((ConnectionEditPart) targetObj).getSource());
                }
            } else {
                targetElem = inverseGraphMap.get(targetObj);
            }
            
            ElkConnectableShape targetShape = null;
            ElkNode targetNode = null;
            ElkPort targetPort = null;
            if (targetElem instanceof ElkNode) {
                targetNode = (ElkNode) targetElem;
                targetShape = targetNode;
            } else if (targetElem instanceof ElkPort) {
                targetPort = (ElkPort) targetElem;
                targetNode = targetPort.getParent();
                targetShape = targetPort;
            } else {
                continue;
            }

            // calculate offset for edge and label coordinates
            ElkNode edgeContainment = ElkGraphUtil.findLowestCommonAncestor(sourceNode, targetNode);
            
            KVector offset = new KVector();
            ElkUtil.toAbsolute(offset, edgeContainment);

            if (!isOppositeEdge) {
                // set source and target
                edge.getSources().add(sourceShape);
                edge.getTargets().add(targetShape);
                
                // now that source and target are set, put the edge into the graph
                edgeContainment.getContainedEdges().add(edge);

                mapping.getGraphMap().put(edge, connection);

                // store the current coordinates of the edge
                setEdgeLayout(edge, connection, offset);
            }

            // process edge labels
            processEdgeLabels(mapping, connection, edge, edgeLabelPlacement, offset);
        }
    }
    
    
    /**
     * Stores the layout information of the given connection edit part into an edge layout.
     * 
     * @param edge
     *            an edge layout
     * @param connection
     *            a connection edit part
     * @param offset
     *            offset to be subtracted from coordinates
     */
    protected void setEdgeLayout(final ElkEdge edge, final ConnectionEditPart connection, final KVector offset) {
        Connection figure = connection.getConnectionFigure();
        PointList pointList = figure.getPoints();

        // our edge will have exactly one edge section
        ElkEdgeSection edgeSection = ElkGraphUtil.createEdgeSection(edge);

        // source point
        Point firstPoint = pointList.getPoint(0);
        edgeSection.setStartX(firstPoint.x - offset.x);
        edgeSection.setStartY(firstPoint.y - offset.y);

        // bend points
        for (int i = 1; i < pointList.size() - 1; i++) {
            Point point = pointList.getPoint(i);
            ElkGraphUtil.createBendPoint(edgeSection, point.x - offset.x, point.y - offset.y);
        }

        // target point
        Point lastPoint = pointList.getPoint(pointList.size() - 1);
        edgeSection.setEndX(lastPoint.x - offset.x);
        edgeSection.setEndY(lastPoint.y - offset.y);

        //
        if (figure instanceof Shape) {
            double currentSize = ((Shape) figure).getLineWidth();
            if (currentSize != CoreOptions.EDGE_THICKNESS.getDefault()) {
                edge.setProperty(CoreOptions.EDGE_THICKNESS, currentSize);
            }
        }
    }
    
    /**
     * Process the labels of an edge.
     * 
     * @param mapping
     *            the layout mapping
     * @param connection
     *            the connection edit part
     * @param edge
     *            the layout edge
     * @param placement
     *            predefined placement for all labels, or {@code Optional#empty()} if the placement shall be derived
     *            from the edit part
     * @param offset
     *            the offset for coordinates
     */
    protected void processEdgeLabels(final LayoutMapping mapping, final ConnectionEditPart connection, final ElkEdge edge, final Optional<EdgeLabelPlacement> placement, final KVector offset) {
        /*
         * ars: source and target is exchanged when defining it in the gmfgen file. So if Emma sets a label to be placed
         * as target on a connection, then the label will show up next to the source node in the diagram editor. So
         * correct it here, very ugly.
         */
        for (Object obj : connection.getChildren()) {
            if (obj instanceof LabelEditPart) {
                LabelEditPart labelEditPart = (LabelEditPart) obj;
                IFigure labelFigure = labelEditPart.getFigure();

                // Check if the label is visible in the first place
                if (labelFigure == null || !labelFigure.isVisible()) {
                    continue;
                }

                Rectangle labelBounds = getAbsoluteBounds(labelFigure);
                String labelText = null;
                Dimension iconBounds = null;

                if (labelFigure instanceof WrappingLabel) {
                    WrappingLabel wrappingLabel = (WrappingLabel) labelFigure;
                    labelText = wrappingLabel.getText();
                    if (wrappingLabel.getIcon() != null) {
                        iconBounds = new Dimension();
                        iconBounds.width = wrappingLabel.getIcon().getBounds().width + wrappingLabel.getIconTextGap();
                        iconBounds.height = wrappingLabel.getIcon().getBounds().height;
                        // Add more characters to the text for layouters that
                        // need the text to
                        // determine the label size.
                        labelText = "O " + labelText;
                    }
                } else if (labelFigure instanceof Label) {
                    Label label = (Label) labelFigure;
                    labelText = label.getText();
                    if (label.getIcon() != null) {
                        iconBounds = label.getIconBounds().getSize();
                        iconBounds.width += label.getIconTextGap();
                        // Add more characters to the text for layouters that
                        // need the text to
                        // determine the label size.
                        labelText = "O " + labelText;
                    }
                } else if (labelFigure instanceof SiriusWrapLabel) {
                    SiriusWrapLabel label = (SiriusWrapLabel) labelFigure;
                    labelText = label.getText();
                    if (label.getIcon() != null) {
                        iconBounds = label.getIconBounds().getSize();
                        iconBounds.width += label.getIconTextGap();
                        // Add more characters to the text for layouters that
                        // need the text to
                        // determine the label size.
                        labelText = "O " + labelText;
                    }
                }

                if (labelText != null && labelText.length() > 0) {
                    ElkLabel label = ElkGraphUtil.createLabel(edge);
                    if (!placement.isPresent()) {
                        switch (labelEditPart.getKeyPoint()) {
                        case ConnectionLocator.SOURCE:
                            label.setProperty(CoreOptions.EDGE_LABELS_PLACEMENT, EdgeLabelPlacement.HEAD);
                            break;
                        case ConnectionLocator.MIDDLE:
                            label.setProperty(CoreOptions.EDGE_LABELS_PLACEMENT, EdgeLabelPlacement.CENTER);
                            break;
                        case ConnectionLocator.TARGET:
                            label.setProperty(CoreOptions.EDGE_LABELS_PLACEMENT, EdgeLabelPlacement.TAIL);
                            break;
                        }
                    } else {
                        label.setProperty(CoreOptions.EDGE_LABELS_PLACEMENT, placement.get());
                    }
                    
                    Font font = labelFigure.getFont();
                    if (font != null && !font.isDisposed()) {
                        label.setProperty(CoreOptions.FONT_NAME, font.getFontData()[0].getName());
                        label.setProperty(CoreOptions.FONT_SIZE, font.getFontData()[0].getHeight());
                    }
                    
                    label.setLocation(labelBounds.x - offset.x, labelBounds.y - offset.y);
                    // The label width includes the icon width
                    label.setWidth(labelBounds.width);
                    label.setHeight(labelBounds.height);

                    // We would set the modified flag to false here, but that
                    // doesn't exist anymore

                    label.setText(labelText);
                    mapping.getGraphMap().put(label, labelEditPart);
                } else {
                    // add the label to the mapping anyway so it is reset to its
                    // reference location
                    ElkLabel label = ElkGraphUtil.createLabel(null);
                    mapping.getGraphMap().put(label, labelEditPart);
                }
            }
        }
    }


    @Override
    public void applyLayout(final LayoutMapping mapping, final IPropertyHolder settings) {
        applyLayout(mapping, 0);
    }
	
	protected void applyLayout(final LayoutMapping mapping) {
        Command applyLayoutCommand = mapping.getProperty(LAYOUT_COMMAND);
        
        if (applyLayoutCommand != null) {
            // Get a command stack to execute the command
            CommandStack commandStack = mapping.getProperty(COMMAND_STACK);
            IWorkbenchPart workbenchPart = mapping.getWorkbenchPart();
            if (commandStack == null) {
                if (workbenchPart != null) {
                    Object adapter = workbenchPart.getAdapter(CommandStack.class);
                    if (adapter instanceof CommandStack) {
                        commandStack = (CommandStack) adapter;
                    }
                }
                if (commandStack == null) {
                    IGraphicalEditPart parentElement = (IGraphicalEditPart) mapping.getParentElement();
                    commandStack = parentElement.getDiagramEditDomain().getDiagramCommandStack();
                }
            }
    
            // Execute the command
            commandStack.execute(applyLayoutCommand);
            
            // Refresh the border items in the diagram
            DiagramEditor diagramEditor = getDiagramEditor(workbenchPart);
            if (diagramEditor != null) {
                refreshDiagram(diagramEditor, (IGraphicalEditPart) mapping.getParentElement());
            }
        }
    }
	
	/**
     * Refreshes all ports in the diagram. This is necessary in order correctly move ports, which
     * does not work due to GMF bugs. See Eclipse bug #291484.
     * 
     * @param editor
     *            the diagram editor
     * @param rootPart
     *            the root edit part
     * @see https://bugs.eclipse.org/bugs/show_bug.cgi?id=291484
     */
    private static void refreshDiagram(final DiagramEditor editor, final IGraphicalEditPart rootPart) {
        EditPart editPart = rootPart;
        if (editPart == null) {
            editPart = editor.getDiagramEditPart();
        }
        for (Object obj : editPart.getViewer().getEditPartRegistry().values()) {
            if (obj instanceof ShapeNodeEditPart) {
                IFigure figure = ((ShapeNodeEditPart) obj).getFigure();
                if (figure instanceof BorderedNodeFigure) {
                    IFigure portContainer = ((BorderedNodeFigure) figure).getBorderItemContainer();
                    portContainer.invalidate();
                    portContainer.validate();
                }
            }
        }
    }
    
    /**
     * Apply the computed layout to the original diagram.
     * 
     * @param mapping a layout mapping that was created by this layout connector
     * @param animationTime the animation time in milliseconds, or 0 for no animation
     */
    protected void applyLayout(final LayoutMapping mapping, final int animationTime) {
        // transfer layout to the diagram
        transferLayout(mapping);
        if (animationTime > 0) {
            // apply the layout with animation
            Animation.markBegin();
            applyLayout(mapping);
            Animation.run(animationTime);
        } else {
            // apply the layout without animation
            applyLayout(mapping);
        }
    }
	
}
