package cz.muni.fi.capella.converter.services;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.util.Properties;

import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.ui.preferences.ScopedPreferenceStore;

import cz.muni.fi.capella.converter.PropertyUtil;
import cz.muni.fi.capella.converter.ui.ConverterUIActivator;
import cz.muni.fi.capella.converter.ui.DialogHandler;

/**
 * This class is responsible for exporting properties into correct file.
 * 
 * @author Andrea Jonásová
 *
 */
public class ExportService {
	
	/**
	 * Creates path to file to which properties should be exported.
	 * 
	 * @return Path to file in format <i>directoryFromPreferences/alg_filename.properties</i>
	 */
	public String createPathForExport() {
		String filename = DialogHandler.getFileNameFromUser();
		
		// if button Cancel was pressed
		if (filename == null) {
			return null;
		}
		
		// get destination directory from preferences settings
        String destinationDir = new ScopedPreferenceStore(InstanceScope.INSTANCE, ConverterUIActivator.PLUGIN_ID)
        		.getString(PropertyUtil.CUSTOM_PROPERTY_DIR_PREF_ID);
		filename = new StringBuilder()
				.append(destinationDir).append(File.separatorChar)
				.append(PropertyUtil.PROPERTY_FILE_PREFIX).append(filename).append(PropertyUtil.PROPERTY_FILE_SUFIX)
				.toString();
		
		if(Files.exists(Path.of(filename), LinkOption.NOFOLLOW_LINKS) 
				&& !DialogHandler.displayOverwriteMessageDialog()) {
				filename = createPathForExport();
		}
		return filename;
	}
	
	/**
	 * Exports properties into given file.
	 * 
	 * @param filename File to which properties should be exported.
	 * @param properties Properties to export.
	 */
	public void exportPropertiesToFile(String filename, Properties properties) {
		try (FileOutputStream outputStream = new FileOutputStream(filename)){
			properties.store(outputStream, null);
		} catch (IOException e) {
			DialogHandler.displayFileErrorDialog(e.getMessage());
		}
	}
	

}
