package cz.muni.fi.capella.converter.ui;

import java.util.regex.Pattern;

import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;

/**
 * This class is responsible for displaying pop up windows.
 *  
 * @author Andrea Jonásová
 *
 */
public final class DialogHandler {

	/**
	 * Displays dialog for user to choose file name.
	 * 
	 * @return File name from user.
	 */
	public static String getFileNameFromUser() {
		InputDialog dialog = new InputDialog(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
				"Export properties", "Please enter file name", "", 
				new IInputValidator() {
					
					@Override
					public String isValid(String input) {
						// only file names containing letters, numbers, _ , - , . are allowed  
						Pattern filePattern = Pattern.compile("[a-zA-Z0-9_\\.-]+");
						if(filePattern.matcher(input).matches()) {
							return null;
						}
						return "Invalid input";
					}
					
				});
		
		dialog.open();
		return dialog.getValue();
	}
	
	/**
	 * Displays error dialog for invalid properties.
	 * 
	 * @param entry Information about part, where error occurred.
	 */
	public static final void displayPropertyErrorDialog(String message) {
		MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), 
				"File format error", 
				"Values could not be loaded. Your file contains one or more errors. \n Error on line: " + message );
	}
	
	/**
	 * Displays error dialog for errors that occurred when manipulating with files.
	 * 
	 * @param fileError Information about error.
	 */
	public static final void displayFileErrorDialog(String fileError) {
		MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), 
				"File error", fileError);
	}
	
	/**
	 * Displays message about overwriting your file, you can choose whether to continue.
	 * 
	 * @return <code>true</code> if file should be overwritten, <code>false</code> otherwise.
	 */
	public static final boolean displayOverwriteMessageDialog() {
		return MessageDialog.openQuestion(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
		"File already exists", "File already exists! Do you want to overwrite it?");
	}
	
	/**
	 * Displays error dialog for unsupported operation with message about operation.
	 * 
	 * @param message Message explaining why was operation unsupported.
	 */
	public static final void displayUnsupportedOperationErrorDialog(String message) {
		MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), 
				"Unsupported operation", message);
	}
}
