package cz.muni.fi.capella.converter.ui.importing;

import java.io.File;
import java.util.Properties;

import javax.naming.OperationNotSupportedException;

import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.elk.core.ui.views.LayoutViewPart;
import org.eclipse.jface.action.Action;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.preferences.ScopedPreferenceStore;

import cz.muni.fi.capella.converter.PropertyUtil;
import cz.muni.fi.capella.converter.exceptions.InvalidPropertyException;
import cz.muni.fi.capella.converter.services.ImportService;
import cz.muni.fi.capella.converter.services.PropertyService;
import cz.muni.fi.capella.converter.ui.ConverterUIActivator;
import cz.muni.fi.capella.converter.ui.DialogHandler;

/**
 * This class is responsible for loading all properties from selected file. 
 * Firstly, all values are checked in order to determine whether file is in valid format. 
 * After check, all values are stored in LayoutView.
 * 
 * @author Andrea Jonásová
 * 
 */
public class LayoutViewAction extends Action {

	private String fileId;
	private boolean isDefault;
	
	private ImportService importService = new ImportService();
	
	public LayoutViewAction(String fileId, boolean isDefault) {
		super(getName(fileId));
		this.fileId = fileId;
		this.isDefault = isDefault;
	}
	
	@Override
	public void run() {
		Properties properties;
		
		// load all properties from selected file
		if (isDefault) {
			properties = importService.loadDefProperties(PropertyUtil.PROP_DIRECTORY_PATH + File.separator + fileId);
		} else {
	        String userSourceDir = new ScopedPreferenceStore(InstanceScope.INSTANCE, ConverterUIActivator.PLUGIN_ID)
	        		.getString(PropertyUtil.CUSTOM_PROPERTY_DIR_PREF_ID);
			properties = importService.loadCustomProperties(userSourceDir + File.separator + fileId);
		}
		
		PropertyService propService = new PropertyService();
		
		try {
			properties = propService.tryConvertProperties(properties);
			// if all properties were valid, store them in property source
			propService.writePropertiesToPropertySource(properties);
		} catch (InvalidPropertyException exception) {
			DialogHandler.displayPropertyErrorDialog(exception.getMessage());
			return;
		} catch (OperationNotSupportedException exception) {
			DialogHandler.displayUnsupportedOperationErrorDialog(exception.getMessage());;
			return;
		}
		
		// refresh layout view
		((LayoutViewPart) PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActivePart())
				.refresh();
	}
	
	/**
	 * Gets name for item in the menu from <code>fileName</code>.
	 * 
	 * @param fileName File name.
	 * @return <code>fileName</code> without prefix and suffix.
	 */
	private static final String getName(String fileName) {
		return fileName.substring(PropertyUtil.PROPERTY_FILE_PREFIX.length(), 
				fileName.length() - PropertyUtil.PROPERTY_FILE_SUFIX.length());
	}
}
