package cz.muni.fi.capella.converter.ui.menu;

import org.eclipse.elk.core.LayoutConfigurator;
import org.eclipse.elk.core.data.LayoutAlgorithmData;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.service.DiagramLayoutEngine;
import org.eclipse.elk.graph.ElkNode;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;

/**
 * This class is responsible for initiating execution of chosen algorithm on diagram.
 * 
 * @author Andrea Jonásová
 */
public class AlgorithmAction extends Action {
	
	private LayoutAlgorithmData algorithmData;
	
	public AlgorithmAction(LayoutAlgorithmData algorithmData) {
		super(algorithmData.getName());
		setToolTipText(algorithmData.getDescription());
		this.algorithmData = algorithmData;
	}

	@Override
	public void run() {
		
		IWorkbenchPart workbench = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActivePart();
		ISelection currentSelection = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getSelectionService()
				.getSelection();
		
		// get selected part
		Object part = null;
		if (currentSelection != null && !currentSelection.isEmpty() && currentSelection instanceof IStructuredSelection) {
		    IStructuredSelection selection = (IStructuredSelection) currentSelection;
		    part = selection.toList();
		}
		
		LayoutConfigurator configurator = new LayoutConfigurator();
		// more properties can be set here -> algorithm will be executed with these when button in context menu is pressed
		configurator.configure(ElkNode.class)
			.setProperty(CoreOptions.ALGORITHM, algorithmData.getId());
		
		DiagramLayoutEngine.Parameters parameters = new DiagramLayoutEngine.Parameters();
		parameters.setOverrideDiagramConfig(true);
		parameters.addLayoutRun(); // must be here to avoid replacing configurator with LayoutView settings
		parameters.addLayoutRun(configurator); 
		
		DiagramLayoutEngine.invokeLayout(workbench, part, parameters);
	}
}
