package cz.muni.fi.capella.converter.ui.menu;

import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

import org.eclipse.elk.core.data.LayoutAlgorithmData;
import org.eclipse.elk.core.data.LayoutMetaDataService;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.ui.actions.CompoundContributionItem;

/**
 * This class is responsible for creating item in context menu for each correctly registered ELK algorithm.
 * 
 * @author Andrea Jonásová
 */
public class DynamicAlgorithmItem extends CompoundContributionItem {

	@Override
	protected IContributionItem[] getContributionItems() {
		// get data for all properly registered algorithms 
		Collection<LayoutAlgorithmData> algorithmData = LayoutMetaDataService.getInstance().getAlgorithmData();
		IContributionItem[] items = new IContributionItem[algorithmData.size()];
		
		// sort algorithms
		List<LayoutAlgorithmData> sortedData = algorithmData.stream().sorted(new Comparator<LayoutAlgorithmData>() {

			@Override
			public int compare(LayoutAlgorithmData data1, LayoutAlgorithmData data2) {
				return data1.getName().compareTo(data2.getName());
			}
			
		}).collect(Collectors.toList());
		
		// convert algorithm names to AlgorithmActions
		int index = 0;
		for (LayoutAlgorithmData data : sortedData) {
			items[index++] = new ActionContributionItem(new AlgorithmAction(data));
		}
		return items;
	}
}
